//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package store

import (
	"fmt"
	"io"
	"os"
	"os/user"
	"path/filepath"
	"reflect"
	"sort"
	"testing"
	"time"

	"github.com/coreos/bbolt"
)

var (
	testBasePath = "run/aura-tank_testing"
	testDBPath   = "/run/aura-tank_testing/.db.bolt"
	testGroup1   = "test1"
	testGroup2   = "test2"
)

func stringInSlice(slice []string, search string) bool {
	for _, element := range slice {
		if element == search {
			return true
		}
	}
	return false
}

func TestMain(m *testing.M) {
	u, err := user.Current()
	if err != nil {
		os.Exit(-1)
	}
	testBasePath = fmt.Sprintf("/run/user/%s/aura-tank_testing", u.Uid)
	testDBPath = filepath.Join(testBasePath, ".db.bolt")
	os.Exit(m.Run())
}

//
// Testing
//

func TestOpen(t *testing.T) {
	// base-path is non-existing directory
	if _, err := NewStore(&Config{"/nonexistend/"}); err == nil {
		t.Fatalf("opening store in nonexisting directory should throw an error")
	}

	// store db-path is a directory
	os.Remove(testDBPath)
	if err := os.MkdirAll(testDBPath, 0700); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if _, err := NewStore(&Config{testBasePath}); err == nil {
		t.Fatalf("opening store using a directory should throw an error")
	}

	// exisiting but non-database file
	os.Remove(testDBPath)
	if f, err := os.Create(testDBPath); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		io.WriteString(f, "this is not a bolt db.")
		f.Close()
	}
	if _, err := NewStore(&Config{testBasePath}); err == nil {
		t.Fatalf("opening store using a invalid database should throw an error")
	}

	// bolt db no version key
	os.Remove(testDBPath)
	if db, err := bolt.Open(testDBPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			if _, err := tx.CreateBucket([]byte(infoBn)); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(&Config{testBasePath}); err == nil {
		t.Fatalf("opening store without a database version should throw an error")
	}

	// bolt db wrong version
	if db, err := bolt.Open(testDBPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			b := tx.Bucket([]byte(infoBn))
			if err := b.Put([]byte(storeVersionKey), itob(0)); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(&Config{testBasePath}); err == nil {
		t.Fatalf("opening store with wrong database version should throw an error")
	}

	// TODO: add checks for
	//    - missing/wrong buckets (ie. store intregrity)

	// create new bolt-db and reopen it
	os.Remove(testDBPath)
	store, err := NewStore(&Config{testBasePath})
	if err != nil {
		t.Fatalf("creating new store failed: %v", err)
	}
	store.Close()

	store, err = NewStore(&Config{testBasePath})
	if err != nil {
		t.Fatalf("re-opening existing store failed: %v", err)
	}
	if _, err := NewStore(&Config{testBasePath}); err == nil {
		t.Fatalf("opening already opened database should throw an error")
	}
	store.Close()
}

// Groups
//

func TestGroups(t *testing.T) {
	os.Remove(testDBPath)
	store, err := NewStore(&Config{testBasePath})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	groups, err := store.ListGroups()
	if err != nil {
		t.Fatalf("listing groups failed: %v", err)
	}
	if len(groups) != 1 && groups[0] != publicGroupBn {
		t.Fatalf("a newly created store should contain a single group '%s' but ListGroups returned: %q", publicGroupBn, groups)
	}

	if err = store.createGroup(testGroup1); err != nil {
		t.Fatalf("creating group failed: %v", err)
	}
	groups, err = store.ListGroups()
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	groupsSorted := sort.StringSlice(groups)
	groupsSorted.Sort()

	groupsExpected := sort.StringSlice([]string{publicGroupBn, testGroup1})
	groupsExpected.Sort()

	if !reflect.DeepEqual(groupsSorted, groupsExpected) {
		t.Fatalf("store should contain groups %q but ListGroups returned: %q", groupsExpected, groupsSorted)
	}

	if err = store.createGroup(testGroup2); err != nil {
		t.Fatalf("creating group failed: %v", err)
	}
	groups, err = store.ListGroups()
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	groupsSorted = sort.StringSlice(groups)
	groupsSorted.Sort()

	groupsExpected = sort.StringSlice([]string{publicGroupBn, testGroup1, testGroup2})
	groupsExpected.Sort()

	if !reflect.DeepEqual(groupsSorted, groupsExpected) {
		t.Fatalf("store should contain groups %q but ListGroups returned: %q", groupsExpected, groupsSorted)
	}
}

// Files
//

func TestFiles(t *testing.T) {
	os.Remove(testDBPath)
	store, err := NewStore(&Config{testBasePath})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	files, err := store.ListFiles(publicGroupBn)
	if err != nil {
		t.Fatalf("listing files of public group failed: %v", err)
	}
	if len(files) != 0 {
		t.Fatalf("a newly created store should contain no files in public group but ListFiles returned: %q", files)
	}

	files, err = store.ListFiles("notexistend")
	if err != nil {
		t.Fatalf("listing files of not existing group shouldn't throw an error but returned: %v", err)
	}
	if len(files) != 0 {
		t.Fatalf("listing files of not existing group should return and empty list but ListFiles returned: %q", files)
	}
}
